//C++
#ifndef CSTAT_H
#define CSTAT_H 1

#include <xsTypes.h>
#include <XSstreams.h>
#include <XSContainer.h>
#include <XSStat/FitStatMethod.h>
#include <XSModel/Data/SpectralData.h>
#include <XSModel/Data/BackCorr/Background.h>
#include <XSModel/GlobalContainer/DataSetTypes.h>
#include <XSModel/GlobalContainer/ModelContainer.h>
#include <XSModel/GlobalContainer/ModelTypes.h>
#include <XSModel/Model/Model.h>
#include <XSUtil/Error/Error.h>
#include <algorithm>
#include <cmath>
#include <map>
#include <utility>
#include <sstream>
#include <numeric>

class Model;

template <typename T>
class Cstat : public FitStatMethod
{
  public:

     Cstat();
     virtual ~Cstat(); 

     virtual Real statValueForSingleBin (Real obs, Real mod, Real error, Real areaTime, const Real& NODATA) const;
     virtual Real deltaSigmaForSingleBin (Real obs, Real mod, Real error, Real areaTime, const Real& NODATA) const;
     virtual Real sumDerivs (const RealArray& dFolded, const int spectrumNumber, 
			     int parameterIndex) const;
     virtual Real sumSecondDerivs (const std::map<size_t,RealArray>& dFolded, int pi, int pj);
     virtual const std::string fullName () const;
     virtual const std::string scriptName () const;
     virtual const std::string commandName () const;

     virtual bool isModifierAllowed () const;

  private:

     Cstat(const Cstat<T> &right);
     Cstat<T>& operator=(const Cstat<T>& right);

     virtual void doReportWarnings (const SpectralData*) const;
     virtual void doPerform (const SpectralData*);
     virtual void doInitialize (const SpectralData*);
     virtual void doDeallocate ();
     virtual void doReset (const SpectralData*);

     virtual RealArray getDifferences (const SpectralData*) const;
     virtual std::pair<Real,Real> adjustForFrozen (const SpectralData* spec, const std::vector<Model*>& models, const std::map<Model*,ArrayContainer >& savedValsForFrozen, const int nFitSpecs);

     bool checkForPoissonSource (const SpectralData*) const;
     bool checkForPoissonBackground (const SpectralData*) const;
     RealArray& firstDerivCoefficient ();
     void firstDerivCoefficient (const RealArray& value);
     RealArray& secondDerivCoefficient ();
     void secondDerivCoefficient (const RealArray& value);

     RealArray m_firstDerivCoefficient;
     RealArray m_secondDerivCoefficient;

};

template <typename T>
Cstat<T>::Cstat()
 : FitStatMethod(T::variantName),
  m_firstDerivCoefficient(),
  m_secondDerivCoefficient()
{
   canUseChainRule(true);
}

template <typename T>
Cstat<T>::~Cstat()
{
}

template <typename T>
void Cstat<T>::doPerform(const SpectralData* sp)
{
  const size_t specNum = sp->spectrumNumber();
  const std::valarray<size_t>& IN = sp->indirectNotice();

  doReset(sp);

  RealArray tmpQ(protoWorkspace());
  if (tmpQ.size() != IN.size()) {
    std::ostringstream oss;
    oss <<"Array size error for spectrum "<< specNum << " in Cstat::doPerform.";
    throw RedAlert(oss.str());
  }

  // for each model,

  XSContainer::ModelMapConstIter ml (XSContainer::models->modelSet().begin());
  XSContainer::ModelMapConstIter mlEnd (XSContainer::models->modelSet().end());
  while (ml != mlEnd) {
    // add up the folded models for each spectrum [ >= indicates multiple sources]
    Model* m (ml->second);
    if (!m->isActive()) {
      ++ml;
      continue;
    }
    ArrayContainer::const_iterator itFold = m->foldedModel().find(specNum);
    if (itFold != m->foldedModel().end())    
       tmpQ += m->foldedModel().find(specNum)->second[IN];
    ++ml;
  }

  // sum contributions.
  // stat is C/2

  Real stat = 0.0;
  RealArray ts (sp->exposureTime(),IN.size());
  RealArray S (ts * sp->spectrum()[IN]);
  const RealArray& model = tmpQ;
  const Background* background (sp->background());
  const Background* correction (sp->correction());

  if (!background) {
    int ifail = T::specificPerform(S,ts,model,number(),modifier(),stat);
    if ( std::isnan(stat) && ifail >= 0 && !std::isnan(model[ifail]) ) {
      tcout << "NaN detected when calculating statistic for bin " 
	    << ifail << " of spectrum " << sp->spectrumNumber() 
	    << std::endl;
      tcout << S[ifail] << " " << " " << ts[ifail] << " " << model[ifail] << std::endl;
      value(2.0*stat);
      return;
    }
  } else {
    // tb includes the background exposure time multiplied by the ratios
    // between the source and background BACKSCALE and AREASCALE.
    RealArray tb (background->data()->backgroundScale()[IN]);
    tb /= sp->backgroundScale()[IN];
    tb *= background->data()->areaScale()[IN];
    tb /= sp->areaScale()[IN];
    tb *= background->data()->exposureTime();
    RealArray B (tb * background->spectrum()[IN]);
    RealArray Berr (tb * background->variance()[IN]);

    if (correction && sp->correctionScale() != 0) {
      Real cTime (correction->data()->exposureTime());
      RealArray tc (correction->data()->backgroundScale()[IN]);
      tc /= sp->backgroundScale()[IN];
      tc *= sp->correctionScale();
      tc *= cTime;
      RealArray C (tc * correction->spectrum()[IN]);
      for (size_t i=0; i<B.size(); i++) B[i] += C[i];
    }

    int ifail = T::specificPerformB(S,B,Berr,ts,tb,model,number(),modifier(),stat);
    if (std::isnan(stat) && ifail >= 0 && 
	!std::isnan(model[ifail]) && !std::isinf(model[ifail]) ) {
      tcout << "NaN detected when calculating statistic for bin " 
	    << ifail << " of spectrum " << sp->spectrumNumber() 
	    << std::endl;
      tcout << S[ifail] << " " << B[ifail] << " " << Berr[ifail] << " " 
	    << ts[ifail] << " " << tb[ifail] << " " << model[ifail] << std::endl;
      value(stat);
      return;
    }
  }

  value(2.0*stat);

}

template <typename T>
void Cstat<T>::doInitialize(const SpectralData* sp)
{
  FitStatMethod::doInitialize(sp);

  // allocate for derivative coefficient arrays.
  const size_t N = sp->indirectNotice().size();
  m_firstDerivCoefficient.resize(N,0.);
  m_secondDerivCoefficient.resize(N,0.);
}

template <typename T>
void Cstat<T>::doDeallocate()
{
    FitStatMethod::doDeallocate();
    m_firstDerivCoefficient.resize(0);
    m_secondDerivCoefficient.resize(0);
}

template <typename T>
void Cstat<T>::doReset(const SpectralData* sp)
{
  const size_t specNum = sp->spectrumNumber();
  const std::valarray<size_t>& IN = sp->indirectNotice();

  RealArray tmpQ(protoWorkspace()); 
  if (tmpQ.size() != IN.size()) {
    std::ostringstream oss;
    oss <<"Array size error for spectrum "<< specNum << " in Cstat::doReset.";
    throw RedAlert(oss.str());
  }

  // for each model,
  XSContainer::ModelMapConstIter itMod (XSContainer::models->modelSet().begin());      
  XSContainer::ModelMapConstIter itModEnd (XSContainer::models->modelSet().end());      
  while (itMod != itModEnd) {
    // add up the folded models for each spectrum [ >= indicates multiple sources]
    Model* m (itMod->second);
    if (!m->isActive()) {
      ++itMod;
      continue;
    }
    ArrayContainer::const_iterator itFold = m->foldedModel().find(specNum);
    if (itFold != m->foldedModel().end())    
       tmpQ += itFold->second[IN];
    ++itMod;
  }    

  RealArray ts (sp->exposureTime(),IN.size());
  RealArray S (ts * sp->spectrum()[IN]);
  // output arrays for derivative coefficients.
  // diff1 is -C'/2 and diff2 is C''/2
  RealArray& diff1 = m_firstDerivCoefficient;
  RealArray& diff2 = m_secondDerivCoefficient;
  RealArray& model = tmpQ;
  const Background* background (sp->background());
  if (!background) {
    T::specificResetCalc(S, ts, model, number(), modifier(), diff1, diff2);
  } else {
    // Variable definitions follow Appendix B of manual.
    // tb includes the background exposure time multiplied by the ratios
    // between the source and background BACKSCALE and AREASCALE.
    RealArray tb (background->data()->backgroundScale()[IN]);
    tb /= sp->backgroundScale()[IN];
    tb *= background->data()->areaScale()[IN];
    tb /= sp->areaScale()[IN];
    tb *= background->data()->exposureTime();
    RealArray B (tb * background->spectrum()[IN]);
    RealArray Berr (tb * background->variance()[IN]);

    T::specificResetCalcB(S, B, Berr, ts, tb, model, number(), modifier(), diff1, diff2);
  }

}
// end doReset

template <typename T>
void Cstat<T>::doReportWarnings (const SpectralData* sp) const
{
  using namespace std;

  bool PoissonSource = checkForPoissonSource(sp);
  bool PoissonBackground = checkForPoissonBackground(sp);

  if (!T::specificValidStatistic(PoissonSource, PoissonBackground)) {
    tcout << "\nWarning: "<< T::variantName <<" statistic is only valid for Poisson data." << std::endl;
    if (!PoissonSource)
      tcout << "    Source file is not Poisson for spectrum " << sp->spectrumNumber() << std::endl;
    if (!PoissonBackground)
      tcout << "    Background file is not Poisson for spectrum " << sp->spectrumNumber() << std::endl;
    tcout << std::endl;
  }
}

template <typename T>
Real Cstat<T>::statValueForSingleBin (Real obs, Real mod, Real error, Real areaTime, const Real& NODATA) const
{
  Real point (0);
  // Presumeably areaTime is not .0 or it wouldn't have made it here.
  if ( obs > 0 )
  {
       obs *= areaTime;
       mod *= areaTime;
       if (mod > 0)
       {
          Real sign = (obs >= mod) ? 1.0 : -1.0;
          point = 2.0*sign*(mod - obs*(1 - (log(obs) - log(mod))));
       }
       else
          point = NODATA;
  }
  else
  {
        point = 2*mod*areaTime;       
  }      
  return point;                    
}

template <typename T>
Real Cstat<T>::deltaSigmaForSingleBin (Real obs, Real mod, Real error, Real areaTime, const Real& NODATA) const
{
  Real oN = obs * areaTime;
  Real mN = mod * areaTime;

  Real point = oN - mN;

  // Use the model as the estimator for the variance unless the model is zero
  // in which case use the data - if that is zero too then use the Gehrels error
  // on zero counts

  Real weight = sqrt(mN);
  if ( weight <= 0.0 ) {
    weight = sqrt(oN);
    if ( weight <= 0.0 ) {
      weight = (1.0 + sqrt(0.75));
    }
  }

  //  tcout << oN << " " << mN << " " << weight << " " << point/weight << std::endl;

  return point/weight;                    
}

template <typename T>
RealArray Cstat<T>::getDifferences (const SpectralData* sp) const
{
  RealArray result;

  const std::valarray<size_t>& IN = sp->indirectNotice();
  const size_t specNum = sp->spectrumNumber();
  result.resize(IN.size());
  result = sp->spectrum()[IN];
  if (sp->background()) {
    result -= sp->background()->spectrum()[IN];
  }

  XSContainer::ModelMapConstIter ml (XSContainer::models->modelSet().begin());      
  XSContainer::ModelMapConstIter  mlEnd (XSContainer::models->modelSet().end()); 
  while ( ml != mlEnd )
  {
     // D - F for each model. In the case where there are multiple
     // sources difference will be iteratively defined.
     Model* m (ml->second);
     if (m->isActive())
     {
       ArrayContainer::const_iterator itFluxEnd = m->modelFlux().end();
       ArrayContainer::const_iterator itFlux = m->modelFlux().find(specNum);
       if ( itFlux != itFluxEnd ) {
	 if ( result.size() != IN.size() ) {
	   std::ostringstream oss;
	   oss << "Array size for spectrum " << specNum << " in Cstat::getDifferences.";
	   throw RedAlert(oss.str());
	 }
	 result -= m->foldedModel(specNum)[IN];
	 if ( tpout.maxChatter() >= 33 ) {
	   RealArray spec (sp->spectrum()[IN]);
           tcout << " difference " << specNum << " source " << m->sourceNumber() << '\n';
           size_t N (result.size());
           for (size_t j = 0; j < N; ++j) {
	     tcout  << j  << " -= " << result[j] << "  " 
		    << "  " << spec[j] << "  " <<  m->foldedModel(specNum)[IN[j]] << '\n';
           }
           tcout << " spcounts " << spec.sum() << " mcounts " << m->foldedModel(specNum)[IN].sum() << '\n';
           tcout << std::flush;
	 }
       }
     }
     ++ml;
  }    
  return result;
}
// end getDifferences

template <typename T>
Real Cstat<T>::sumDerivs (const RealArray& dFolded, const int spectrumNumber, 
			  const int parameterIndex) const
{

    //dFolded = derivatives for input parameter.
    // See comments in ChiSquare::sumDerivs for a detailed explanation.

    Real dSdpj(0);

    RealArray dSdpA (m_firstDerivCoefficient);
    dSdpA *= dFolded;
    dSdpj += -2.0*dSdpA.sum();
    if (false) {
      size_t N = m_firstDerivCoefficient.size();
      tcout << N << std::endl;
      std::vector<Real> PS(N);
      std::partial_sum(&dSdpA[0],&dSdpA[0]+N,PS.begin());
      tcout << " Spectrum " << spectrumNumber << " Parameter# " << parameterIndex  
	    << " D:  " << dSdpj << '\n';
      for (size_t j = 0; j < N; ++j) {
	tcout << j << " " << std::setw(12) 
	      << std::setw(12) << m_firstDerivCoefficient[j] << " "
	      << std::setw(12) << dFolded[j]  << " "
	      << std::setw(12) << dSdpA[j] << " "
	      << std::setw(12) << PS[j]  << '\n'; 
      }
      tcout << " d " << spectrumNumber << "  " << dSdpj << '\n';  
    }

    return dSdpj;

}
// end sumDerivs

template <typename T>
Real Cstat<T>::sumSecondDerivs (const std::map<size_t,RealArray>& dFolded, int pi, int pj)
{
   const RealArray& piDeriv = dFolded.find(pi)->second;
   const RealArray& pjDeriv = dFolded.find(pj)->second;

   RealArray secondDeriv(m_secondDerivCoefficient);
   secondDeriv *= piDeriv;
   secondDeriv *= pjDeriv;
   Real alpha_ij = 2.0*secondDeriv.sum();
   return alpha_ij;
} 
// end sumSecondDerivs

template <typename T>
std::pair<Real,Real> Cstat<T>::adjustForFrozen (const SpectralData* spec, const std::vector<Model*>& models, const std::map<Model*,ArrayContainer >& savedValsForFrozen, const int nFitSpecs)
{

    const std::valarray<size_t>& IN = spec->indirectNotice();
    const size_t N = IN.size();
    const size_t specNum = spec->spectrumNumber();

    std::pair<Real,Real> fSums (0.,0.);

    RealArray data (spec->spectrum()[IN]);
    RealArray ts (spec->exposureTime(), N);
    RealArray totModFroz(0., N); 
    RealArray totalModel(0., N);

    for (size_t i=0; i<models.size(); ++i)
    {
      Model* mod = models[i];
      std::map<Model*,ArrayContainer >::const_iterator itSaved =
                savedValsForFrozen.find(mod);
      if (itSaved != savedValsForFrozen.end())
      {
         totalModel += itSaved->second.find(specNum)->second[IN];
         totModFroz += mod->foldedModel(specNum)[IN];
      }
      else
      {
         totalModel += mod->foldedModel(specNum)[IN];
      }
    }

    fSums.first = ((data - totModFroz)*ts).sum();
    fSums.second = ((totalModel - totModFroz)*ts).sum();

   return fSums;
}

template <typename T>
bool Cstat<T>::checkForPoissonSource (const SpectralData* sp) const
{
  return sp->isPoisson();
}

template <typename T>
bool Cstat<T>::checkForPoissonBackground (const SpectralData* sp) const
{
   if (sp->background()) {
     if (!sp->background()->data()->isPoisson()) return false;
   }
   return true;
}


template <typename T>
inline RealArray& Cstat<T>::firstDerivCoefficient ()
{
  return m_firstDerivCoefficient;
}

template <typename T>
inline void Cstat<T>::firstDerivCoefficient (const RealArray& value)
{
  m_firstDerivCoefficient = value;
}

template <typename T>
inline RealArray& Cstat<T>::secondDerivCoefficient ()
{
  return m_secondDerivCoefficient;
}

template <typename T>
inline void Cstat<T>::secondDerivCoefficient (const RealArray& value)
{
  m_secondDerivCoefficient = value;
}

template <typename T>
inline const std::string Cstat<T>::fullName () const
{
  return T::specificFullName(number(), modifier());
}

template <typename T>
inline const std::string Cstat<T>::scriptName () const
{
  return T::specificScriptName(number(), modifier());
}

template <typename T>
inline const std::string Cstat<T>::commandName () const
{
  return T::specificCommandName(number(), modifier());
}

template <typename T>
inline bool Cstat<T>::isModifierAllowed () const
{
  return true;
}

#endif
